use std::{marker::PhantomData, str::FromStr};

use serde::{
    de::{DeserializeOwned, Visitor},
    Deserialize, Deserializer, Serialize,
};
use time::OffsetDateTime;

use crate::error::{Error, Result as MxmResult};

//#COMMON

#[derive(Debug, Deserialize)]
pub struct Resp<T> {
    pub message: T,
}

#[derive(Debug, Deserialize)]
pub struct HeaderMsg {
    pub header: Header,
}

#[derive(Debug, Deserialize)]
pub struct BodyMsg<T> {
    pub body: T,
}

#[derive(Debug, Deserialize)]
pub struct Header {
    pub status_code: u16,
    /// Short error message
    #[serde(default)]
    pub hint: String,
}

pub fn parse_body<T: DeserializeOwned>(response: &str) -> MxmResult<T> {
    let header = serde_json::from_str::<Resp<HeaderMsg>>(response)?
        .message
        .header;
    if header.status_code < 400 {
        let body = serde_json::from_str::<Resp<BodyMsg<T>>>(response)?;
        Ok(body.message.body)
    } else if header.status_code == 404 {
        Err(Error::NotFound)
    } else if header.status_code == 401 && header.hint == "renew" {
        Err(Error::TokenExpired)
    } else if header.status_code == 401 && header.hint == "captcha" {
        Err(Error::Ratelimit)
    } else {
        Err(Error::MusixmatchError {
            status_code: header.status_code,
            msg: header.hint,
        })
    }
}

//#SESSION

#[derive(Debug, Deserialize)]
pub struct GetToken {
    pub user_token: String,
}

//#LOGIN

#[derive(Debug, Serialize)]
pub struct Credentials<'a> {
    pub credential_list: &'a [CredentialWrap<'a>],
}

#[derive(Debug, Serialize)]
pub struct CredentialWrap<'a> {
    pub credential: Credential<'a>,
}

#[derive(Debug, Serialize)]
pub struct Credential<'a> {
    pub action: &'a str,
    pub email: &'a str,
    pub password: &'a str,
    #[serde(rename = "type")]
    pub typ: &'a str,
}

#[derive(Debug, Deserialize)]
pub struct Login(pub Vec<LoginInner>);

#[derive(Debug, Deserialize)]
pub struct LoginInner {
    pub credential: LoginCredential,
}

#[derive(Debug, Deserialize)]
#[serde(untagged)]
pub enum LoginCredential {
    Account { account: Account },
    Error { error: ErrorMxm },
}

#[derive(Debug, Deserialize)]
pub struct Account {
    pub id: String,
    pub email: String,
    pub name: String,
}

#[derive(Debug, Deserialize)]
pub struct ErrorMxm {
    pub status_code: u16,
    #[serde(default)]
    pub description: String,
}

pub fn bool_from_int<'de, D>(deserializer: D) -> Result<bool, D::Error>
where
    D: Deserializer<'de>,
{
    struct BoolFromIntVisitor;

    impl<'de> Visitor<'de> for BoolFromIntVisitor {
        type Value = bool;

        fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
            formatter.write_str("a number or boolean")
        }

        fn visit_bool<E>(self, v: bool) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v)
        }

        fn visit_u8<E>(self, v: u8) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_u16<E>(self, v: u16) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_u32<E>(self, v: u32) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_u128<E>(self, v: u128) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_i8<E>(self, v: i8) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_i16<E>(self, v: i16) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_i32<E>(self, v: i32) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_i64<E>(self, v: i64) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_i128<E>(self, v: i128) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0)
        }

        fn visit_f32<E>(self, v: f32) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0.0)
        }

        fn visit_f64<E>(self, v: f64) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v != 0.0)
        }
    }

    deserializer.deserialize_any(BoolFromIntVisitor)
}

/**
 * The Musixmatch API returns zero as a default value for numeric IDs.
 * These values should be deserialized as [`None`].
 */
pub fn null_if_zero<'de, D, N>(deserializer: D) -> Result<Option<N>, D::Error>
where
    D: Deserializer<'de>,
    N: TryFrom<u64>,
{
    struct NullIfZeroVisitor<N> {
        n: PhantomData<N>,
    }

    impl<'de, N> Visitor<'de> for NullIfZeroVisitor<N>
    where
        N: TryFrom<u64>,
    {
        type Value = Option<N>;

        fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
            formatter.write_str("unsigned int or None")
        }

        fn visit_none<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_unit<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            if v < 1 {
                Ok(None)
            } else {
                Ok(v.try_into().ok())
            }
        }

        fn visit_u32<E>(self, v: u32) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }

        fn visit_u16<E>(self, v: u16) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }

        fn visit_u8<E>(self, v: u8) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }
    }

    deserializer.deserialize_any(NullIfZeroVisitor { n: PhantomData })
}

/// The Musixmatch API returns an empty string as a default value for string fields.
/// These values should be deserialized as [`None`].
pub fn null_if_empty<'de, D>(deserializer: D) -> Result<Option<String>, D::Error>
where
    D: Deserializer<'de>,
{
    struct NullIfEmptyVisitor;

    impl<'de> Visitor<'de> for NullIfEmptyVisitor {
        type Value = Option<String>;

        fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
            formatter.write_str("string or None")
        }

        fn visit_none<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_unit<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            if v.is_empty() {
                Ok(None)
            } else {
                Ok(Some(v.to_owned()))
            }
        }
    }

    deserializer.deserialize_any(NullIfEmptyVisitor)
}

/// Deserialize a numeric string into an integer.
/// Return None if the string is empty/null or if there was a parse error.
pub fn parse_int<'de, D, N>(deserializer: D) -> Result<Option<N>, D::Error>
where
    D: Deserializer<'de>,
    N: FromStr + TryFrom<u64>,
{
    struct ParseIntVisitor<N> {
        n: PhantomData<N>,
    }

    impl<'de, N> Visitor<'de> for ParseIntVisitor<N>
    where
        N: FromStr + TryFrom<u64>,
    {
        type Value = Option<N>;

        fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
            formatter.write_str("numeric string or None")
        }

        fn visit_none<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_unit<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(v.try_into().ok())
        }

        fn visit_u32<E>(self, v: u32) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }

        fn visit_u16<E>(self, v: u16) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }

        fn visit_u8<E>(self, v: u8) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            self.visit_u64(v.into())
        }

        fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            if v.is_empty() {
                Ok(None)
            } else {
                Ok(v.parse().ok())
            }
        }
    }

    deserializer.deserialize_any(ParseIntVisitor { n: PhantomData })
}

pub mod optional_date {
    use super::*;
    use serde::ser::Error as _;
    use serde::Serializer;
    use time::{macros::format_description, Date};

    const DATE_FORMAT: &[time::format_description::FormatItem] =
        format_description!("[year]-[month]-[day]");

    pub fn serialize<S: Serializer>(
        value: &Option<Date>,
        serializer: S,
    ) -> Result<S::Ok, S::Error> {
        match value {
            Some(date) => date
                .format(&DATE_FORMAT)
                .map_err(S::Error::custom)?
                .serialize(serializer),
            None => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D: Deserializer<'de>>(
        deserializer: D,
    ) -> Result<Option<Date>, D::Error> {
        struct OptionalDateVisitor;

        impl<'de> Visitor<'de> for OptionalDateVisitor {
            type Value = Option<Date>;

            fn expecting(
                &self,
                formatter: &mut serde::__private::fmt::Formatter,
            ) -> serde::__private::fmt::Result {
                formatter.write_str("date or empty string")
            }

            fn visit_none<E>(self) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                Ok(None)
            }

            fn visit_unit<E>(self) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                Ok(None)
            }

            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                if v.is_empty() || v == "0000-00-00" {
                    Ok(None)
                } else {
                    Ok(Date::parse(v, &DATE_FORMAT).ok())
                }
            }
        }

        deserializer.deserialize_any(OptionalDateVisitor)
    }
}

pub mod optional_datetime {
    use super::*;
    use serde::Serializer;
    use time::format_description::well_known::Rfc3339;

    pub fn serialize<S: Serializer>(
        value: &Option<OffsetDateTime>,
        serializer: S,
    ) -> Result<S::Ok, S::Error> {
        match value {
            Some(datetime) => time::serde::rfc3339::serialize(datetime, serializer),
            None => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D: Deserializer<'de>>(
        deserializer: D,
    ) -> Result<Option<OffsetDateTime>, D::Error> {
        struct OptionalDateVisitor;

        impl<'de> Visitor<'de> for OptionalDateVisitor {
            type Value = Option<OffsetDateTime>;

            fn expecting(
                &self,
                formatter: &mut serde::__private::fmt::Formatter,
            ) -> serde::__private::fmt::Result {
                formatter.write_str("timestamp or empty string")
            }

            fn visit_none<E>(self) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                Ok(None)
            }

            fn visit_unit<E>(self) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                Ok(None)
            }

            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                if v.is_empty() {
                    Ok(None)
                } else {
                    OffsetDateTime::parse(v, &Rfc3339)
                        .map_err(E::custom)
                        .map(Some)
                }
            }
        }

        deserializer.deserialize_any(OptionalDateVisitor)
    }
}

#[cfg(test)]
mod tests {
    use time::Date;

    use super::*;
    use crate::models::subtitle::SubtitleBody;

    #[test]
    fn deserialize_no_body() {
        let json =
            r#"{"message":{"header":{"status_code":401,"execute_time":0.002,"hint":"fsck"}}}"#;

        let err = parse_body::<SubtitleBody>(json).unwrap_err();

        if let Error::MusixmatchError { status_code, msg } = err {
            assert_eq!(status_code, 401);
            assert_eq!(msg, "fsck");
        } else {
            panic!("invalid error: {err}");
        }
    }

    #[test]
    fn deserialize_body() {
        let json = r#"{"message":{"header":{"status_code":200,"execute_time":0.002},"body":"Hello World"}}"#;

        let res = parse_body::<String>(json).unwrap();
        assert_eq!(res, "Hello World");
    }

    #[test]
    fn deserialize_null_if_zero() {
        #[derive(Deserialize)]
        struct S {
            #[serde(default, deserialize_with = "null_if_zero")]
            n: Option<u64>,
        }

        let json = r#"{"n": 0}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, None);

        let json = r#"{"n": 1}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, Some(1));

        let json = r#"{"n": null}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, None);

        let json = r#"{}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, None);
    }

    #[test]
    fn deserialize_null_if_empty() {
        #[derive(Deserialize)]
        struct S {
            #[serde(default, deserialize_with = "null_if_empty")]
            s: Option<String>,
        }

        let json = r#"{"s": ""}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().s, None);

        let json = r#"{"s": "a"}"#;
        assert_eq!(
            serde_json::from_str::<S>(json).unwrap().s,
            Some("a".to_owned())
        );

        let json = r#"{"n": null}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().s, None);

        let json = r#"{}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().s, None);
    }

    #[test]
    fn deserialize_parse_int() {
        #[derive(Deserialize)]
        struct S {
            #[serde(default, deserialize_with = "parse_int")]
            n: Option<u64>,
        }

        let json = r#"{"n": 0}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, Some(0));

        let json = r#"{"n": "0"}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, Some(0));

        let json = r#"{"n": null}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, None);

        let json = r#"{}"#;
        assert_eq!(serde_json::from_str::<S>(json).unwrap().n, None);
    }

    #[test]
    fn deserialize_optional_date() {
        #[derive(Deserialize)]
        struct S {
            #[serde(with = "optional_date")]
            date: Option<Date>,
        }

        let json_null_string = r#"{"date": null}"#;
        let json_empty_string = r#"{"date": ""}"#;
        let json_zero_date = r#"{"date": "0000-00-00"}"#;
        let json_date = r#"{"date": "2022-08-27"}"#;

        let res = serde_json::from_str::<S>(json_null_string).unwrap();
        assert!(res.date.is_none());

        let res = serde_json::from_str::<S>(json_empty_string).unwrap();
        assert!(res.date.is_none());

        let res = serde_json::from_str::<S>(json_zero_date).unwrap();
        assert!(res.date.is_none());

        let res = serde_json::from_str::<S>(json_date).unwrap();
        assert!(res.date.is_some());
    }

    #[test]
    fn deserialize_optional_datetime() {
        #[derive(Deserialize)]
        struct S {
            #[serde(with = "optional_datetime")]
            date: Option<OffsetDateTime>,
        }

        let json_null_string = r#"{"date": null}"#;
        let json_empty_string = r#"{"date": ""}"#;
        let json_date = r#"{"date": "2022-08-27T23:47:20Z"}"#;

        let res = serde_json::from_str::<S>(json_null_string).unwrap();
        assert!(res.date.is_none());

        let res = serde_json::from_str::<S>(json_empty_string).unwrap();
        assert!(res.date.is_none());

        let res = serde_json::from_str::<S>(json_date).unwrap();
        assert!(res.date.is_some());
    }
}
