use crate::error::Result;
use crate::models::album::{Album, AlbumBody, AlbumListBody};
use crate::models::{AlbumId, ArtistId, SortOrder};
use crate::Musixmatch;

impl Musixmatch {
    /// Get the metadata for an album specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Album ID](crate::models::AlbumId)
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/album-get>
    pub async fn album(&self, id: AlbumId<'_>) -> Result<Album> {
        let mut url = self.new_url("album.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.finish();
        }

        let album_body = self.execute_get_request::<AlbumBody>(&url).await?;
        Ok(album_body.album)
    }

    /// Get the album discography of an artist specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Artist ID](crate::models::ArtistId)
    /// - `s_release_date`: Sort the albums by release date. If None, the albums are
    ///   sorted by popularity
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/artist-albums-get>
    pub async fn artist_albums(
        &self,
        artist_id: ArtistId<'_>,
        s_release_date: Option<SortOrder>,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Album>> {
        let mut url = self.new_url("artist.albums.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = artist_id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            if let Some(s_release_date) = s_release_date {
                url_query.append_pair("s_release_date", s_release_date.as_str());
            }
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let album_list_body = self.execute_get_request::<AlbumListBody>(&url).await?;
        Ok(album_list_body
            .album_list
            .into_iter()
            .map(|a| a.album)
            .collect())
    }

    /// This api provides you the list of the top albums of a given country.
    ///
    /// # Parameters
    /// - `country`: A valid country code (default: "US")
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// None - please update your documentation, Musixmatch!
    pub async fn chart_albums(
        &self,
        country: &str,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Album>> {
        let mut url = self.new_url("chart.albums.get");
        {
            let mut url_query = url.query_pairs_mut();

            url_query.append_pair("country", country);
            url_query.append_pair("chart_name", "new_releases");
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let album_list_body = self.execute_get_request::<AlbumListBody>(&url).await?;
        Ok(album_list_body
            .album_list
            .into_iter()
            .map(|a| a.album)
            .collect())
    }
}
