use crate::error::Result;
use crate::models::artist::{Artist, ArtistBody, ArtistListBody};
use crate::models::ArtistId;
use crate::Musixmatch;

impl Musixmatch {
    /// Get the metadata for an artist specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Artist ID](crate::models::ArtistId)
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/artist-get>
    pub async fn artist(&self, id: ArtistId<'_>) -> Result<Artist> {
        let mut url = self.new_url("artist.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.finish();
        }

        let artist_body = self.execute_get_request::<ArtistBody>(&url).await?;
        Ok(artist_body.artist)
    }

    /// Get a list of artists somehow related to the one specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Artist ID](crate::models::ArtistId)
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/artist-related-get>
    pub async fn artist_related(
        &self,
        id: ArtistId<'_>,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Artist>> {
        let mut url = self.new_url("artist.related.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let artist_list_body = self.execute_get_request::<ArtistListBody>(&url).await?;
        Ok(artist_list_body
            .artist_list
            .into_iter()
            .map(|a| a.artist)
            .collect())
    }

    /// Search for artists in the Musixmatch database.
    ///
    /// # Parameters
    /// - `q_artist`: Search term
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/artist-search>
    pub async fn artist_search(
        &self,
        q_artist: &str,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Artist>> {
        let mut url = self.new_url("artist.search");
        {
            let mut url_query = url.query_pairs_mut();

            url_query.append_pair("q_artist", q_artist);
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let artist_list_body = self.execute_get_request::<ArtistListBody>(&url).await?;
        Ok(artist_list_body
            .artist_list
            .into_iter()
            .map(|a| a.artist)
            .collect())
    }

    /// This api provides you the list of the top artists of a given country.
    ///
    /// # Parameters
    /// - `country`: A valid country code (default: "US")
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/artist-chart-get>
    pub async fn chart_artists(
        &self,
        country: &str,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Artist>> {
        let mut url = self.new_url("chart.artists.get");
        {
            let mut url_query = url.query_pairs_mut();

            if !country.is_empty() {
                url_query.append_pair("country", country);
            }
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let artist_list_body = self.execute_get_request::<ArtistListBody>(&url).await?;
        Ok(artist_list_body
            .artist_list
            .into_iter()
            .map(|a| a.artist)
            .collect())
    }
}
