use crate::error::Result;
use crate::models::lyrics::{Lyrics, LyricsBody};
use crate::models::translation::{TranslationList, TranslationListBody};
use crate::models::TrackId;
use crate::Musixmatch;

impl Musixmatch {
    /// Get the lyrics for a track specified by its name and artist.
    ///
    /// # Parameters
    /// - `q_track`: Title of the track
    /// - `q_artist`: Artist of the track
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/matcher-lyrics-get>
    pub async fn matcher_lyrics(&self, q_track: &str, q_artist: &str) -> Result<Lyrics> {
        let mut url = self.new_url("matcher.lyrics.get");
        {
            let mut url_query = url.query_pairs_mut();
            if !q_track.is_empty() {
                url_query.append_pair("q_track", q_track);
            }
            if !q_artist.is_empty() {
                url_query.append_pair("q_artist", q_artist);
            }
            url_query.finish();
        }

        let lyrics_body = self.execute_get_request::<LyricsBody>(&url).await?;
        lyrics_body.validate()?;
        Ok(lyrics_body.lyrics)
    }

    /// Get the lyrics for a track specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Track ID](crate::models::TrackId)
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/track-lyrics-get>
    pub async fn track_lyrics(&self, id: TrackId<'_>) -> Result<Lyrics> {
        let mut url = self.new_url("track.lyrics.get");
        {
            let mut url_query = url.query_pairs_mut();
            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.finish();
        }

        let lyrics_body = self.execute_get_request::<LyricsBody>(&url).await?;
        lyrics_body.validate()?;
        Ok(lyrics_body.lyrics)
    }

    /// Get translated lyrics for a track specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Track ID](crate::models::TrackId)
    /// - `selected_language`: The language of the translated lyrics
    ///   [(ISO 639‑1)](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes)
    /// # Reference
    /// None, the [public translation API](https://developer.musixmatch.com/documentation/api-reference/track-lyrics-translation-get)
    /// is only available on commercial plans
    pub async fn track_lyrics_translation(
        &self,
        id: TrackId<'_>,
        selected_language: &str,
    ) -> Result<TranslationList> {
        let mut url = self.new_url("crowd.track.translations.get");
        {
            let mut url_query = url.query_pairs_mut();
            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.append_pair("translation_fields_set", "minimal");
            url_query.append_pair("selected_language", selected_language);
            url_query.finish();
        }

        let translation_list_body = self
            .execute_get_request::<TranslationListBody>(&url)
            .await?;
        Ok(TranslationList {
            lines: translation_list_body
                .translations_list
                .into_iter()
                .map(|t| t.translation)
                .collect(),
            lang: selected_language.to_owned(),
        })
    }
}
