use crate::error::Result;
use crate::models::subtitle::{Subtitle, SubtitleBody, SubtitleFormat};
use crate::models::TrackId;
use crate::Musixmatch;

impl Musixmatch {
    /// Get the subtitles (synchronized lyrics) for a track specified by its name and artist.
    ///
    /// # Parameters
    /// - `q_track`: Title of the track
    /// - `q_artist`: Artist of the track
    /// - `subtitle_format`: [Format](crate::models::SubtitleFormat) of the subtitles
    /// - `f_subtitle_length`: Optional subtitle length (track duration) in seconds
    /// - `f_subtitle_length_max_deviation`: Optional maximum amount of seconds the subtitle length
    ///   is allowed to deviate from the given value. The Musixmatch app sets this value to 1,
    ///   so this should be the recommended value.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/matcher-subtitle-get>
    pub async fn matcher_subtitle(
        &self,
        q_track: &str,
        q_artist: &str,
        subtitle_format: SubtitleFormat,
        f_subtitle_length: Option<f32>,
        f_subtitle_length_max_deviation: Option<f32>,
    ) -> Result<Subtitle> {
        let mut url = self.new_url("matcher.subtitle.get");
        {
            let mut url_query = url.query_pairs_mut();
            if !q_track.is_empty() {
                url_query.append_pair("q_track", q_track);
            }
            if !q_artist.is_empty() {
                url_query.append_pair("q_artist", q_artist);
            }
            url_query.append_pair("subtitle_format", subtitle_format.to_param());
            if let Some(f_subtitle_length) = f_subtitle_length {
                url_query.append_pair("f_subtitle_length", &f_subtitle_length.to_string());
            }
            if let Some(f_subtitle_length_max_deviation) = f_subtitle_length_max_deviation {
                url_query.append_pair(
                    "f_subtitle_length_max_deviation",
                    &f_subtitle_length_max_deviation.to_string(),
                );
            }
            if subtitle_format == SubtitleFormat::TtmlStructured {
                url_query.append_pair("part", "track_structure");
            }
            url_query.finish();
        }

        let subtitle_body = self.execute_get_request::<SubtitleBody>(&url).await?;
        Ok(subtitle_body.subtitle)
    }

    /// Get the subtitles (synchronized lyrics) for a track specified by its ID.
    ///
    /// # Parameters
    /// - `id`: [Track ID](crate::models::TrackId)
    /// - `subtitle_format`: [Format](crate::models::SubtitleFormat) of the subtitles
    /// - `f_subtitle_length`: Optional subtitle length (track duration) in seconds
    /// - `f_subtitle_length_max_deviation`: Optional maximum amount of seconds the subtitle length
    ///   is allowed to deviate from the given value. The Musixmatch app sets this value to 1,
    ///   so this should be the recommended value.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/track-subtitle-get>
    pub async fn track_subtitle(
        &self,
        id: TrackId<'_>,
        subtitle_format: SubtitleFormat,
        f_subtitle_length: Option<f32>,
        f_subtitle_length_max_deviation: Option<f32>,
    ) -> Result<Subtitle> {
        let mut url = self.new_url("track.subtitle.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            url_query.append_pair("subtitle_format", subtitle_format.to_param());
            if let Some(f_subtitle_length) = f_subtitle_length {
                url_query.append_pair("f_subtitle_length", &f_subtitle_length.to_string());
            }
            if let Some(f_subtitle_length_max_deviation) = f_subtitle_length_max_deviation {
                url_query.append_pair(
                    "f_subtitle_length_max_deviation",
                    &f_subtitle_length_max_deviation.to_string(),
                );
            }
            if subtitle_format == SubtitleFormat::TtmlStructured {
                url_query.append_pair("part", "track_structure");
            }
            url_query.finish();
        }

        let subtitle_body = self.execute_get_request::<SubtitleBody>(&url).await?;
        Ok(subtitle_body.subtitle)
    }
}
