use time::Date;

use crate::error::Result;
use crate::models::track::{Track, TrackBody, TrackListBody};
use crate::models::{AlbumId, ChartName, Genre, Genres, SortOrder, TrackId};
use crate::Musixmatch;

impl Musixmatch {
    /// Get the metadata for a track specified by its name, artist and album.
    ///
    /// # Parameters
    /// - `q_track`: Title of the track
    /// - `q_artist`: Artist of the track
    /// - `q_album`: Album name of the track
    /// - `translation_status`: Get the lyrics [translation status](crate::models::TrackLyricsTranslationStatus)
    /// - `lang_3c`: Output the lyrics translation status with
    ///   [ISO 639‑2](https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes) 3-letter language codes
    ///   instead of [ISO 639‑1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) codes.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/matcher-track-get>
    pub async fn matcher_track(
        &self,
        q_track: &str,
        q_artist: &str,
        q_album: &str,
        translation_status: bool,
        lang_3c: bool,
    ) -> Result<Track> {
        let mut url = self.new_url("matcher.track.get");
        {
            let mut url_query = url.query_pairs_mut();

            if !q_track.is_empty() {
                url_query.append_pair("q_track", q_track);
            }
            if !q_artist.is_empty() {
                url_query.append_pair("q_artist", q_artist);
            }
            if !q_album.is_empty() {
                url_query.append_pair("q_album", q_album);
            }
            if translation_status {
                url_query.append_pair("part", "track_lyrics_translation_status");
                url_query.append_pair(
                    "language_iso_code",
                    match lang_3c {
                        true => "0",
                        false => "1",
                    },
                );
            }
            url_query.finish();
        }

        let track_body = self.execute_get_request::<TrackBody>(&url).await?;
        Ok(track_body.track)
    }

    /// Get the metadata for a track specified by its id.
    ///
    /// # Parameters
    /// - `id`: [Track ID](crate::models::TrackId)
    /// - `translation_status`: Get the lyrics [translation status](crate::models::TrackLyricsTranslationStatus)
    /// - `lang_3c`: Output the lyrics translation status with
    ///   [ISO 639‑2](https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes) 3-letter language codes
    ///   instead of [ISO 639‑1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) codes.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/track-get>
    pub async fn track(
        &self,
        id: TrackId<'_>,
        translation_status: bool,
        lang_3c: bool,
    ) -> Result<Track> {
        let mut url = self.new_url("track.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            if translation_status {
                url_query.append_pair("part", "track_lyrics_translation_status");
                url_query.append_pair(
                    "language_iso_code",
                    match lang_3c {
                        true => "0",
                        false => "1",
                    },
                );
            }
            url_query.finish();
        }

        let track_body = self.execute_get_request::<TrackBody>(&url).await?;
        Ok(track_body.track)
    }

    /// Get the list of songs of an album.
    ///
    /// # Parameters
    /// - `id`: [Album ID](crate::models::AlbumId)
    /// - `f_has_lyrics`: When true, filter only contents with lyrics
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/album-tracks-get>
    pub async fn album_tracks(
        &self,
        id: AlbumId<'_>,
        f_has_lyrics: bool,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Track>> {
        let mut url = self.new_url("album.tracks.get");
        {
            let mut url_query = url.query_pairs_mut();

            let id_param = id.to_param();
            url_query.append_pair(id_param.0, &id_param.1);
            if f_has_lyrics {
                url_query.append_pair("f_has_lyrics", "1");
            }
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let track_list_body = self.execute_get_request::<TrackListBody>(&url).await?;
        Ok(track_list_body
            .track_list
            .into_iter()
            .map(|t| t.track)
            .collect())
    }

    /// Get a list of the top songs of a given country.
    ///
    /// # Parameters
    /// - `country`: A valid country code (default: "US")
    /// - `chart_name`: Select among [available charts](crate::models::ChartName)
    /// - `f_has_lyrics`: When true, filter only contents with lyrics
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/track-chart-get>
    pub async fn chart_tracks(
        &self,
        country: &str,
        chart_name: ChartName,
        f_has_lyrics: bool,
        page_size: u8,
        page: u32,
    ) -> Result<Vec<Track>> {
        let mut url = self.new_url("chart.tracks.get");
        {
            let mut url_query = url.query_pairs_mut();

            if !country.is_empty() {
                url_query.append_pair("country", country);
            }
            url_query.append_pair("chart_name", chart_name.as_str());
            if f_has_lyrics {
                url_query.append_pair("f_has_lyrics", "1");
            }
            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let track_list_body = self.execute_get_request::<TrackListBody>(&url).await?;
        Ok(track_list_body
            .track_list
            .into_iter()
            .map(|t| t.track)
            .collect())
    }

    /// Get the list of the music genres the Musixmatch catalogue.
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/music-genres-get>
    pub async fn genres(&self) -> Result<Vec<Genre>> {
        let url = self.new_url("music.genres.get");
        let genres = self.execute_get_request::<Genres>(&url).await?;
        Ok(genres.music_genre_list)
    }

    /// Create a new query builder for searching tracks in the Musixmatch database.
    ///
    /// **Note:** The search results are unordered the by default. You probably want
    /// to sort by popularity (`.s_track_rating(SortOrder::Desc)`) to get relevant results.
    ///
    /// # Example
    /// ```ignore
    /// let tracks = musixmatch
    ///   .track_search()
    ///   .q_lyrics("Never gonna run around and desert you")
    ///   .s_track_rating(SortOrder::Desc)
    ///   .send(10, 1)
    ///   .await?;
    /// ```
    ///
    /// # Reference
    /// <https://developer.musixmatch.com/documentation/api-reference/track-search>
    pub fn track_search(&self) -> TrackSearchQuery {
        TrackSearchQuery {
            mxm: self.clone(),
            q_track: None,
            q_artist: None,
            q_lyrics: None,
            q_track_artist: None,
            q_writer: None,
            q: None,
            f_artist_id: None,
            f_music_genre_id: None,
            f_lyrics_language: None,
            f_has_lyrics: false,
            f_track_release_group_first_release_date_min: None,
            f_track_release_group_first_release_date_max: None,
            s_artist_rating: None,
            s_track_rating: None,
            quorum_factor: None,
        }
    }
}

pub struct TrackSearchQuery<'a> {
    mxm: Musixmatch,

    q_track: Option<&'a str>,
    q_artist: Option<&'a str>,
    q_lyrics: Option<&'a str>,
    q_track_artist: Option<&'a str>,
    q_writer: Option<&'a str>,
    q: Option<&'a str>,
    f_artist_id: Option<u64>,
    f_music_genre_id: Option<u64>,
    f_lyrics_language: Option<&'a str>,
    f_has_lyrics: bool,
    f_track_release_group_first_release_date_min: Option<Date>,
    f_track_release_group_first_release_date_max: Option<Date>,
    s_artist_rating: Option<SortOrder>,
    s_track_rating: Option<SortOrder>,
    quorum_factor: Option<f32>,
}

impl<'a> TrackSearchQuery<'a> {
    /// Track title
    pub fn q_track(mut self, q_track: &'a str) -> Self {
        self.q_track = Some(q_track);
        self
    }

    /// Track artist
    pub fn q_artist(mut self, q_artist: &'a str) -> Self {
        self.q_artist = Some(q_artist);
        self
    }

    /// Any word in the lyrics
    pub fn q_lyrics(mut self, q_lyrics: &'a str) -> Self {
        self.q_lyrics = Some(q_lyrics);
        self
    }

    /// Any word in the song title or artist name
    pub fn q_track_artist(mut self, q_track_artist: &'a str) -> Self {
        self.q_track_artist = Some(q_track_artist);
        self
    }

    /// Track lyrics writer
    pub fn q_writer(mut self, q_writer: &'a str) -> Self {
        self.q_writer = Some(q_writer);
        self
    }

    /// Any word in the song title or artist name or lyrics
    pub fn q(mut self, q: &'a str) -> Self {
        self.q = Some(q);
        self
    }

    /// When set, filter by this artist ID
    pub fn f_artist_id(mut self, f_artist_id: u64) -> Self {
        self.f_artist_id = Some(f_artist_id);
        self
    }

    /// When set, filter by this genre ID
    pub fn f_music_genre_id(mut self, f_music_genre_id: u64) -> Self {
        self.f_music_genre_id = Some(f_music_genre_id);
        self
    }

    /// Filter by the lyrics language ("en", "it", ...)
    pub fn f_lyrics_language(mut self, f_lyrics_language: &'a str) -> Self {
        self.f_lyrics_language = Some(f_lyrics_language);
        self
    }

    /// When set, filter only contents with lyrics
    pub fn f_has_lyrics(mut self) -> Self {
        self.f_has_lyrics = true;
        self
    }

    /// When set, filter the tracks with release date newer than the given value
    pub fn f_track_release_date_min(mut self, f_track_release_date_min: Date) -> Self {
        self.f_track_release_group_first_release_date_min = Some(f_track_release_date_min);
        self
    }

    /// When set, filter the tracks with release date older than the given value
    pub fn f_track_release_date_max(mut self, f_track_release_date_max: Date) -> Self {
        self.f_track_release_group_first_release_date_max = Some(f_track_release_date_max);
        self
    }

    /// Sort by artist popularity (asc|desc)
    pub fn s_artist_rating(mut self, s_artist_rating: SortOrder) -> Self {
        self.s_artist_rating = Some(s_artist_rating);
        self
    }

    /// Sort by track popularity (asc|desc)
    pub fn s_track_rating(mut self, s_track_rating: SortOrder) -> Self {
        self.s_track_rating = Some(s_track_rating);
        self
    }

    /// Search only a part of the given query string.
    ///
    /// Allowed range is (0.1 – 0.9)
    pub fn quorum_factor(mut self, quorum_factor: f32) -> Self {
        self.quorum_factor = Some(quorum_factor);
        self
    }

    /// Execute the search query.
    ///
    /// # Parameters
    /// - `page_size`: Define the page size for paginated results. Range is 1 to 100.
    /// - `page`: Define the page number for paginated results, starting from 1.
    pub async fn send(&self, page_size: u8, page: u32) -> Result<Vec<Track>> {
        let mut url = self.mxm.new_url("track.search");
        {
            let mut url_query = url.query_pairs_mut();

            if let Some(q_track) = self.q_track {
                url_query.append_pair("q_track", q_track);
            }
            if let Some(q_artist) = self.q_artist {
                url_query.append_pair("q_artist", q_artist);
            }
            if let Some(q_lyrics) = self.q_lyrics {
                url_query.append_pair("q_lyrics", q_lyrics);
            }
            if let Some(q_track_artist) = self.q_track_artist {
                url_query.append_pair("q_track_artist", q_track_artist);
            }
            if let Some(q_writer) = self.q_writer {
                url_query.append_pair("q_writer", q_writer);
            }
            if let Some(q) = self.q {
                url_query.append_pair("q", q);
            }
            if let Some(f_artist_id) = self.f_artist_id {
                url_query.append_pair("f_artist_id", &f_artist_id.to_string());
            }
            if let Some(f_music_genre_id) = self.f_music_genre_id {
                url_query.append_pair("f_music_genre_id", &f_music_genre_id.to_string());
            }
            if let Some(f_lyrics_language) = self.f_lyrics_language {
                url_query.append_pair("f_lyrics_language", f_lyrics_language);
            }
            if self.f_has_lyrics {
                url_query.append_pair("f_has_lyrics", "1");
            }
            if let Some(f_track_release_group_first_release_date_min) =
                self.f_track_release_group_first_release_date_min
            {
                if let Ok(date) =
                    f_track_release_group_first_release_date_min.format(crate::YMD_FORMAT)
                {
                    url_query.append_pair("f_track_release_group_first_release_date_min", &date);
                }
            }
            if let Some(f_track_release_group_first_release_date_max) =
                self.f_track_release_group_first_release_date_max
            {
                if let Ok(date) =
                    f_track_release_group_first_release_date_max.format(crate::YMD_FORMAT)
                {
                    url_query.append_pair("f_track_release_group_first_release_date_max", &date);
                }
            }
            if let Some(s_artist_rating) = &self.s_artist_rating {
                url_query.append_pair("s_artist_rating", s_artist_rating.as_str());
            }
            if let Some(s_track_rating) = &self.s_track_rating {
                url_query.append_pair("s_track_rating", s_track_rating.as_str());
            }
            if let Some(quorum_factor) = self.quorum_factor {
                url_query.append_pair("quorum_factor", &quorum_factor.to_string());
            }

            url_query.append_pair("page_size", &page_size.to_string());
            url_query.append_pair("page", &page.to_string());
            url_query.finish();
        }

        let track_list_body = self.mxm.execute_get_request::<TrackListBody>(&url).await?;
        Ok(track_list_body
            .track_list
            .into_iter()
            .map(|t| t.track)
            .collect())
    }
}
