use serde::{Deserialize, Serialize};
use time::{Date, OffsetDateTime};

use super::Genres;

#[derive(Debug, Deserialize)]
pub(crate) struct AlbumBody {
    pub album: Album,
}

#[derive(Debug, Deserialize)]
pub(crate) struct AlbumListBody {
    pub album_list: Vec<AlbumBody>,
}

/// Album: an album of songs in the Musixmatch database.
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub struct Album {
    /// Unique Musixmatch Album ID
    pub album_id: u64,
    /// Musicbrainz album ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_mbid: Option<String>,
    /// Album name
    pub album_name: String,
    /// Popularity of the album from 0 to 100
    #[serde(default)]
    pub album_rating: u8,
    /// Number of tracks on the album
    pub album_track_count: u16,
    /// Album release date
    #[serde(default, with = "crate::api_model::optional_date")]
    pub album_release_date: Option<Date>,
    /// Album type
    #[serde(default)]
    pub album_release_type: AlbumType,

    /// Musixmatch artist ID
    pub artist_id: u64,
    /// Artist name
    pub artist_name: String,

    /// List of primary genres
    #[serde(default)]
    pub primary_genres: Genres,
    /// Secondary genres / Subgenres
    ///
    /// Example: primary_genres: Pop, secondary_genres: K-Pop / Mandopop
    #[serde(default)]
    pub secondary_genres: Genres,

    /// Album copyright text
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_copyright: Option<String>,
    /// Album label / recording company
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_label: Option<String>,
    /// Human-readable URL-safe Album ID
    ///
    /// Example: `LMFAO/Party-Rock-5`
    pub album_vanity_id: String,

    /// Date and time when the album was last updated
    #[serde(with = "time::serde::rfc3339")]
    pub updated_time: OffsetDateTime,

    /// Album cover URL (100x100px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_100x100: Option<String>,
    /// Album cover URL (350x350px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_350x350: Option<String>,
    /// Album cover URL (500x500px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_500x500: Option<String>,
    /// Album cover URL (800x800px)
    ///
    /// **Note:** not present on a lot of albums
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_800x800: Option<String>,
}

/// Album type
///
/// Source: <https://developer.musixmatch.com/documentation/music-meta-data>
#[allow(missing_docs)]
#[derive(Default, Copy, Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub enum AlbumType {
    #[default]
    Album,
    Ep,
    Single,
    Compilation,
    Remix,
    Live,
}
