use serde::{Deserialize, Serialize};
use time::{Date, OffsetDateTime};

use super::Genres;

#[derive(Debug, Deserialize)]
pub(crate) struct ArtistBody {
    pub artist: Artist,
}

#[derive(Debug, Deserialize)]
pub(crate) struct ArtistListBody {
    pub artist_list: Vec<ArtistBody>,
}

/// Artist: an artist in the Musixmatch database.
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub struct Artist {
    /// Musixmatch Artist ID
    pub artist_id: u64,
    /// Musicbrainz Artist ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_mbid: Option<String>,

    /// Artist name
    pub artist_name: String,
    /// Artist name in different languages
    #[serde(default)]
    pub artist_name_translation_list: Vec<ArtistNameTranslation>,
    /// Artist origin as a 2-letter country code (e.g. "US")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_country: Option<String>,
    /// Alternative names for the artist (e.g. in different languages)
    #[serde(default)]
    pub artist_alias_list: Vec<ArtistAlias>,
    /// Popularity of the artist from 0 to 100
    #[serde(default)]
    pub artist_rating: u8,

    /// Primary genres of the artist
    #[serde(default)]
    pub primary_genres: Genres,
    /// Secondary genres / Subgenres
    ///
    /// Example: primary_genres: Pop, secondary_genres: K-Pop / Mandopop
    ///
    /// Note that this schema is not applied to all tracks on Musixmatch.
    /// There are for example many K-Pop tracks with both Pop and K-Pop
    /// tagged as primary genre and this field empty.
    #[serde(default)]
    pub secondary_genres: Genres,

    /// URL of the artist's Twitter profile
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_twitter_url: Option<String>,
    /// URL of the artist's website
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_website_url: Option<String>,
    /// URL of the artist's Instagram profile
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_instagram_url: Option<String>,
    /// URL of the artist's TikTok profile
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_tiktok_url: Option<String>,
    /// URL of the artist's Facebook profile
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_facebook_url: Option<String>,

    /// URL-safe human-readable artist ID
    ///
    /// Example: "aespa"
    pub artist_vanity_id: String,
    /// Date and time when the artist was last updated
    #[serde(with = "time::serde::rfc3339")]
    pub updated_time: OffsetDateTime,
    /// Start year of the artist's presence
    #[serde(default, deserialize_with = "crate::api_model::parse_int")]
    pub begin_date_year: Option<u16>,
    /// Start date of the artist's presence
    #[serde(default, with = "crate::api_model::optional_date")]
    pub begin_date: Option<Date>,
    /// End year of the artist's presence
    #[serde(default, deserialize_with = "crate::api_model::parse_int")]
    pub end_date_year: Option<u16>,
    /// End date of the artist's presence
    #[serde(default, with = "crate::api_model::optional_date")]
    pub end_date: Option<Date>,
}

/// Alternative artist name (e.g. different languages)
#[derive(Clone, Debug, PartialEq, Eq, Default, Serialize, Deserialize)]
pub struct ArtistAlias {
    /// Alternative artist name
    pub artist_alias: String,
}

/// Artist name in another language
#[derive(Clone, Debug, PartialEq, Eq, Default, Serialize, Deserialize)]
pub struct ArtistNameTranslation {
    /// Artist name in another language
    pub artist_name_translation: ArtistNameTranslationInner,
}

/// Alternative artist name (e.g. different languages)
#[derive(Clone, Debug, PartialEq, Eq, Default, Serialize, Deserialize)]
pub struct ArtistNameTranslationInner {
    /// Language code (e.g. "EN")
    ///
    /// **Note:** the language code is uppercase for some reason
    pub language: String,
    /// Translated name
    pub translation: String,
}
