use std::borrow::Cow;

/// Track identifiers from different sources
#[derive(Debug, Clone, PartialEq, Eq)]
#[non_exhaustive]
pub enum TrackId<'a> {
    /// Musixmatch ID that is the same for equivalent tracks (e.g. same track on different albums)
    Commontrack(u64),
    /// Unique Musixmatch track ID
    TrackId(u64),
    /// Human-readable Musixmatch ID
    ///
    /// Used in the URLs on the Musixmatch website.
    ///
    /// Example: `aespa/Black-Mamba`
    CommontrackVanity(Cow<'a, str>),
    /// [International Standard Recording Code](https://en.wikipedia.org/wiki/International_Standard_Recording_Code)
    Isrc(Cow<'a, str>),
    /// Musicbrainz ID
    ///
    /// **Note:** Musicbrainz IDs are not present for most tracks in
    /// the Musixmatch database, so use them only with a fallback.
    Musicbrainz(Cow<'a, str>),
    /// Spotify track ID
    Spotify(Cow<'a, str>),
}

impl<'a> TrackId<'a> {
    pub(crate) fn to_param(&self) -> (&'static str, String) {
        match self {
            TrackId::Commontrack(id) => ("commontrack_id", id.to_string()),
            TrackId::TrackId(id) => ("track_id", id.to_string()),
            TrackId::CommontrackVanity(id) => ("commontrack_vanity_id", id.to_string()),
            TrackId::Isrc(id) => ("track_isrc", id.to_string()),
            TrackId::Musicbrainz(id) => ("track_mbid", id.to_string()),
            TrackId::Spotify(id) => ("track_spotify_id", id.to_string()),
        }
    }
}

/// Artist identifiers from different sources
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum ArtistId<'a> {
    /// Unique Musixmatch artist ID
    ArtistId(u64),
    /// Musicbrainz ID
    ///
    /// **Note:** Musicbrainz IDs are not present for most artists in
    /// the Musixmatch database, so use them only with a fallback.
    Musicbrainz(&'a str),
}

impl<'a> ArtistId<'a> {
    pub(crate) fn to_param(&self) -> (&'static str, String) {
        match self {
            ArtistId::ArtistId(id) => ("artist_id", id.to_string()),
            ArtistId::Musicbrainz(id) => ("artist_mbid", id.to_string()),
        }
    }
}

/// Album identifiers from different sources
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum AlbumId<'a> {
    /// Unique Musixmatch artist ID
    AlbumId(u64),
    /// Musicbrainz ID
    ///
    /// **Note:** Musicbrainz IDs are not present for most artists in
    /// the Musixmatch database, so use them only with a fallback.
    Musicbrainz(&'a str),
}

impl<'a> AlbumId<'a> {
    pub(crate) fn to_param(&self) -> (&'static str, String) {
        match self {
            AlbumId::AlbumId(id) => ("album_id", id.to_string()),
            AlbumId::Musicbrainz(id) => ("album_mbid", id.to_string()),
        }
    }
}

/// Sort order for items returned by the Musixmatch API
pub enum SortOrder {
    /// Ascending
    Asc,
    /// Descending
    Desc,
}

impl SortOrder {
    pub(crate) fn as_str(&self) -> &str {
        match self {
            SortOrder::Asc => "asc",
            SortOrder::Desc => "desc",
        }
    }
}
