use serde::{Deserialize, Serialize};
use time::OffsetDateTime;

use crate::{error::Result, Error};

#[derive(Debug, Deserialize)]
pub(crate) struct LyricsBody {
    pub lyrics: Lyrics,
}

impl LyricsBody {
    pub(crate) fn validate(&self) -> Result<()> {
        if self.lyrics.lyrics_body.is_empty()
            && self
                .lyrics
                .lyrics_copyright
                .as_deref()
                .map(|c| c.contains("not authorized"))
                .unwrap_or_default()
        {
            Err(Error::NotAvailable)
        } else {
            Ok(())
        }
    }
}

/// Lyrics from the Musixmatch database.
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Lyrics {
    /// Unique Musixmatch lyrics ID
    pub lyrics_id: u64,
    /// True if the track is instrumental
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub instrumental: bool,
    /// True if the lyrics contain explicit language
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub explicit: bool,
    /// Lyrics text
    pub lyrics_body: String,
    /// Language code (e.g. "en")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub lyrics_language: Option<String>,
    /// Language name (e.g. "English")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub lyrics_language_description: Option<String>,
    /// Copyright text of the lyrics
    ///
    /// Ends with a newline.
    ///
    /// Example:
    /// ```text
    /// Writer(s): David Hodges
    /// Copyright: Emi Blackwood Music Inc., 12.06 Publishing, Hipgnosis Sfh I Limited, Hifi Music Ip Issuer L.p.
    /// ```
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub lyrics_copyright: Option<String>,
    /// Date and time when the lyrics were last updated
    #[serde(with = "time::serde::rfc3339")]
    pub updated_time: OffsetDateTime,
}
