use serde::{Deserialize, Serialize};
use std::fmt::Write;
use time::OffsetDateTime;

use crate::{error::Result, Error};

#[derive(Debug, Deserialize)]
pub(crate) struct SubtitleBody {
    pub subtitle: Subtitle,
}

/// Subtitle formats provided by Musixmatch
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[non_exhaustive]
pub enum SubtitleFormat {
    /// [LRC format](https://en.wikipedia.org/wiki/LRC_(file_format))
    ///
    /// Example:
    /// ```text
    /// [00:05.23] Eyes, in the sky, gazing far into the night
    /// [00:11.42] I raise my hand to the fire, but it's no use
    /// [00:16.57] 'Cause you can't stop it from shining through
    /// [00:21.30] It's true
    /// [00:23.00] Baby let the light shine through
    /// ```
    Lrc,
    /// [TTML/DFXP format](https://en.wikipedia.org/wiki/Timed_Text_Markup_Language)
    ///
    /// Example:
    /// ```xml
    /// <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    /// <tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml" xml:lang="en">
    /// <body>
    ///   <div xml:lang="en">
    ///     <p begin="00:00:05.230" end="00:00:11.420">Eyes, in the sky, gazing far into the night</p>
    ///     <p begin="00:00:11.420" end="00:00:16.570">I raise my hand to the fire, but it's no use</p>
    ///     <p begin="00:00:16.570" end="00:00:21.300">'Cause you can't stop it from shining through</p>
    ///     <p begin="00:00:21.300" end="00:00:23.000">It's true</p>
    ///     <p begin="00:00:23.000" end="00:00:26.610">Baby let the light shine through</p>
    ///   </div>
    /// </body>
    /// </tt>
    /// ```
    Ttml,
    /// [TTML/DFXP format](https://en.wikipedia.org/wiki/Timed_Text_Markup_Language) with
    /// track structure tags
    ///
    /// Song parts:
    /// - `intro`: This is located at the beginning of the song, it can have a different melody
    ///   from both verses and chorus.
    /// - `verse`: Roughly corresponds to a poetic stanza. Generally, there are multiple verses in
    ///   a song, and they usually have different lyrics even though the melody will likely be the same.
    /// - `pre-chorus`: Optional section, included between `verse` and `chorus`.
    /// - `chorus`: Contains the main theme of the song, often containing the song title.
    ///   It is repeated after each verse.
    /// - `bridge`: Most commonly the connection between the last chorus and the previous one.
    ///   It could be very similar to other sections but will most likely provide a shift in mood.
    /// - `hook`: A very repetitive and catchy part of the song, usually its most memorable component.
    ///   Not to be mistaken with the chorus.
    /// - `outro`: Comes at the very end of the song, taking it to its conclusion.
    ///   It can be a repeated part of another section, gradually fading away.
    ///
    /// Example:
    /// ```xml
    /// <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    /// <tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml" xml:lang="en">
    /// <body>
    ///   <div xml:lang="en">
    ///     <div song-part= \"chorus\">
    ///       <p begin="00:00:05.230" end="00:00:11.420">Eyes, in the sky, gazing far into the night</p>
    ///       <p begin="00:00:11.420" end="00:00:16.570">I raise my hand to the fire, but it's no use</p>
    ///       <p begin="00:00:16.570" end="00:00:21.300">'Cause you can't stop it from shining through</p>
    ///       <p begin="00:00:21.300" end="00:00:23.000">It's true</p>
    ///       <p begin="00:00:23.000" end="00:00:26.610">Baby let the light shine through</p>
    ///     </div>
    ///   </div>
    /// </body>
    /// </tt>
    /// ```
    TtmlStructured,
    /// Musixmatch's own JSON format.
    ///
    /// Can be deserialized with the [Subtitle::to_lines] / [SubtitleLines::try_from] method.
    ///
    /// Example:
    /// ```json
    /// [
    ///   {
    ///     "text": "Eyes, in the sky, gazing far into the night",
    ///     "time": { "total": 5.23, "minutes": 0, "seconds": 5, "hundredths": 23 }
    ///   },
    ///   {
    ///     "text": "I raise my hand to the fire, but it's no use",
    ///     "time": { "total": 11.42, "minutes": 0, "seconds": 11, "hundredths": 42 }
    ///   },
    ///   {
    ///     "text": "'Cause you can't stop it from shining through",
    ///     "time": { "total": 16.57, "minutes": 0, "seconds": 16, "hundredths": 57 }
    ///   },
    ///   {
    ///     "text": "It's true",
    ///     "time": { "total": 21.3, "minutes": 0, "seconds": 21, "hundredths": 30 }
    ///   },
    ///   {
    ///     "text": "Baby let the light shine through",
    ///     "time": { "total": 23, "minutes": 0, "seconds": 23, "hundredths": 0 }
    ///   }
    /// ]
    /// ```
    Json,
    /// [EBU-STL](http://fileformats.archiveteam.org/wiki/EBU_STL) (hex-encoded binary format)
    ///
    /// This is a very old subtitle format used for PAL television broadcasting.
    ///
    /// Example (raw):
    /// ```text
    /// 38353053544c32352e303130303030307368696e6520286f726967696e616c206d6978292020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020207368696e6520286f726967
    /// 696e616c206d6978292020202020202020202020206d757369786d6174636820202020202020202020
    /// 2020202020202020202020206d757369786d617463682e636f6d202020202020202020202020202020
    /// 2020206d786d2d33343437343332312020202032323039333032323039333030313030303338303030
    /// 3338303030393932313130303030303030303030303030303030313149544120202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202020
    /// 2020202020202020202020202020202020202020202020202020202020202020202020202020202000
    /// 0100FF000000050000000b00120200457965732c20696e2074686520736b792c2067617a696e672066
    /// 617220696e746f20746865206e696768748f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f
    /// 8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f8f
    /// 8f8f8f8f
    /// ```
    ///
    /// Example (decoded):
    ///
    /// ```text
    /// $ shine.stledu | xxd -r -p | hexyl
    /// ┌────────┬─────────────────────────┬─────────────────────────┬────────┬────────┐
    /// │00000000│ 38 35 30 53 54 4c 32 35 ┊ 2e 30 31 30 30 30 30 30 │850STL25┊.0100000│
    /// │00000010│ 73 68 69 6e 65 20 28 6f ┊ 72 69 67 69 6e 61 6c 20 │shine (o┊riginal │
    /// │00000020│ 6d 69 78 29 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │mix)    ┊        │
    /// │00000030│ 20 20 20 20 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │        ┊        │
    /// │*       │                         ┊                         │        ┊        │
    /// │00000070│ 73 68 69 6e 65 20 28 6f ┊ 72 69 67 69 6e 61 6c 20 │shine (o┊riginal │
    /// │00000080│ 6d 69 78 29 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │mix)    ┊        │
    /// │00000090│ 6d 75 73 69 78 6d 61 74 ┊ 63 68 20 20 20 20 20 20 │musixmat┊ch      │
    /// │000000a0│ 20 20 20 20 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │        ┊        │
    /// │000000b0│ 6d 75 73 69 78 6d 61 74 ┊ 63 68 2e 63 6f 6d 20 20 │musixmat┊ch.com  │
    /// │000000c0│ 20 20 20 20 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │        ┊        │
    /// │000000d0│ 6d 78 6d 2d 33 34 34 37 ┊ 34 33 32 31 20 20 20 20 │mxm-3447┊4321    │
    /// │000000e0│ 32 32 30 39 33 30 32 32 ┊ 30 39 33 30 30 31 30 30 │22093022┊09300100│
    /// │000000f0│ 30 33 38 30 30 30 33 38 ┊ 30 30 30 39 39 32 31 31 │03800038┊00099211│
    /// │00000100│ 30 30 30 30 30 30 30 30 ┊ 30 30 30 30 30 30 30 30 │00000000┊00000000│
    /// │00000110│ 31 31 49 54 41 20 20 20 ┊ 20 20 20 20 20 20 20 20 │11ITA   ┊        │
    /// │00000120│ 20 20 20 20 20 20 20 20 ┊ 20 20 20 20 20 20 20 20 │        ┊        │
    /// │*       │                         ┊                         │        ┊        │
    /// │00000400│ 00 01 00 ff 00 00 00 05 ┊ 00 00 00 0b 00 12 02 00 │0•0×000•┊000•0••0│
    /// │00000410│ 45 79 65 73 2c 20 69 6e ┊ 20 74 68 65 20 73 6b 79 │Eyes, in┊ the sky│
    /// │00000420│ 2c 20 67 61 7a 69 6e 67 ┊ 20 66 61 72 20 69 6e 74 │, gazing┊ far int│
    /// │00000430│ 6f 20 74 68 65 20 6e 69 ┊ 67 68 74 8f 8f 8f 8f 8f │o the ni┊ght×××××│
    /// │00000440│ 8f 8f 8f 8f 8f 8f 8f 8f ┊ 8f 8f 8f 8f 8f 8f 8f 8f │××××××××┊××××××××│
    /// ```
    EbuStl,
}

impl SubtitleFormat {
    pub(crate) fn to_param(self) -> &'static str {
        match self {
            SubtitleFormat::Lrc => "lrc",
            SubtitleFormat::Ttml | SubtitleFormat::TtmlStructured => "dfxp",
            SubtitleFormat::Json => "mxm",
            SubtitleFormat::EbuStl => "stledu",
        }
    }
}

/// Subtitle: a synchronized lyrics subtitle in the Musixmatch database.
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Subtitle {
    /// Numeric Musixmatch subtitle ID
    pub subtitle_id: u64,
    /// Subtitle / synchronized lyrics in the requested format
    pub subtitle_body: String,
    /// Language code (e.g. "en")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub subtitle_language: Option<String>,
    /// Language name (e.g. "English")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub subtitle_language_description: Option<String>,
    /// Copyright text of the lyrics
    ///
    /// Ends with a newline.
    ///
    /// Example:
    /// ```text
    /// Writer(s): David Hodges
    /// Copyright: Emi Blackwood Music Inc., 12.06 Publishing, Hipgnosis Sfh I Limited, Hifi Music Ip Issuer L.p.
    /// ```
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub lyrics_copyright: Option<String>,
    /// Duration of the subtitle in seconds
    pub subtitle_length: u32,
    /// Date and time when the subtitle was last updated
    #[serde(with = "time::serde::rfc3339")]
    pub updated_time: OffsetDateTime,
}

impl Subtitle {
    /// Deserialize the json-formatted subtitles into a SubtitleLines object
    ///
    /// Only works with [SubtitleFormat::Json].
    pub fn to_lines(&self) -> Result<SubtitleLines> {
        Ok(SubtitleLines {
            lines: serde_json::from_str(&self.subtitle_body)?,
            lang: self.subtitle_language.to_owned(),
            length: self.subtitle_length,
        })
    }
}

/// List of subtitle lines
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SubtitleLines {
    /// List of subtitle lines
    pub lines: Vec<SubtitleLine>,
    /// Language code (e.g. "en")
    pub lang: Option<String>,
    /// Duration of the subtitle in seconds
    pub length: u32,
}

impl TryFrom<Subtitle> for SubtitleLines {
    type Error = Error;

    fn try_from(subtitle: Subtitle) -> Result<Self> {
        subtitle.to_lines()
    }
}

impl SubtitleLines {
    /// Convert subtitles into the [JSON](SubtitleFormat::Json) format
    pub fn to_json(&self) -> Result<String> {
        serde_json::to_string(&self).map_err(Error::from)
    }

    /// Convert subtitles into the [LRC](SubtitleFormat::Lrc) format
    pub fn to_lrc(&self) -> String {
        let mut t = self.lines.iter().fold(String::new(), |mut acc, line| {
            _ = writeln!(acc, "[{}] {}", line.time.format_lrc(), line.text);
            acc
        });
        t.pop(); // trailing newline
        t
    }

    /// Convert subtitles into the [Ttml](SubtitleFormat::Ttml) format
    pub fn to_ttml(&self) -> String {
        let mut ttml = format!(
            r#"<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml" xml:lang="{0}">
<body>
  <div xml:lang="{0}">"#,
            self.lang.as_deref().unwrap_or_default(),
        );

        for i in 0..self.lines.len() {
            let line = &self.lines[i];
            let t_next = self
                .lines
                .get(i + 1)
                .map(|l| l.time)
                .unwrap_or_else(|| SubtitleTime::from_s(self.length));

            ttml += &format!(
                "\n    <p begin=\"{}\" end=\"{}\">{}</p>",
                line.time.format_ttml(),
                t_next.format_ttml(),
                escape_xml(&line.text),
            );
        }

        ttml += r#"
  </div>
</body>
</tt>"#;
        ttml
    }
}

/// Replace all html control characters to make a string safe for inserting into HTML.
fn escape_xml(input: &str) -> String {
    let mut buf = String::with_capacity(input.len());

    for c in input.chars() {
        match c {
            '<' => buf.push_str("&lt;"),
            '>' => buf.push_str("&gt;"),
            '&' => buf.push_str("&amp;"),
            '"' => buf.push_str("&quot;"),
            '\'' => buf.push_str("&#x27;"),
            '\n' => buf.push_str("<br>"),
            _ => buf.push(c),
        };
    }
    buf
}

/// Single subtitle line
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct SubtitleLine {
    /// Subtitle line text
    pub text: String,
    /// Subtitle line position
    pub time: SubtitleTime,
}

/// Position of a subtitle line in the track
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct SubtitleTime {
    /// Minute component of the timestamp
    pub minutes: u32,
    /// Second component of the timestamp
    pub seconds: u32,
    /// 1/100 second component of the timestamp
    pub hundredths: u32,
}

impl SubtitleTime {
    fn from_s(s: u32) -> Self {
        let minutes = (s as f32 / 60.0).floor() as u32;
        let seconds = s - minutes * 60;

        Self {
            minutes,
            seconds,
            hundredths: 0,
        }
    }

    /// Get the position of the subtitle line in milliseconds
    pub fn total_ms(&self) -> u32 {
        self.minutes * 60000 + self.seconds * 1000 + self.hundredths * 10
    }

    fn format_lrc(&self) -> String {
        format!(
            "{:02}:{:02}:{:02}",
            self.minutes, self.seconds, self.hundredths
        )
    }

    fn format_ttml(&self) -> String {
        let hours = (self.minutes as f32 / 60.0).floor() as u32;
        let minutes = self.minutes - hours * 60;

        format!(
            "{:02}:{:02}:{:02}.{:02}0",
            hours, minutes, self.seconds, self.hundredths
        )
    }
}
