use serde::{Deserialize, Serialize};
use time::OffsetDateTime;

use super::Genres;

#[derive(Debug, Deserialize)]
pub(crate) struct TrackBody {
    pub track: Track,
}

#[derive(Debug, Deserialize)]
pub(crate) struct TrackListBody {
    pub track_list: Vec<TrackBody>,
}

/// Track: a song in the Musixmatch database
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Track {
    /// Unique Musixmatch Track ID
    pub track_id: u64,
    /// Track Musicbrainz ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub track_mbid: Option<String>,
    /// [International Standard Recording Code](https://en.wikipedia.org/wiki/International_Standard_Recording_Code)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub track_isrc: Option<String>,
    /// [ISRCs](https://en.wikipedia.org/wiki/International_Standard_Recording_Code) of equivalent tracks (e.g. on different albums)
    #[serde(default)]
    pub commontrack_isrcs: Vec<Vec<String>>,
    /// Track ID on Spotify
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub track_spotify_id: Option<String>,
    /// Spotify IDs of equivalent tracks (e.g. on different albums)
    #[serde(default)]
    pub commontrack_spotify_ids: Vec<String>,
    /// Track ID on Soundcloud
    #[serde(default, deserialize_with = "crate::api_model::null_if_zero")]
    pub track_soundcloud_id: Option<u64>,
    /// Track ID on XBox Music
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub track_xboxmusic_id: Option<String>,

    /// Title of the track
    pub track_name: String,
    /// Popularity of the track from 0 to 100
    #[serde(default)]
    pub track_rating: u8,
    /// Length of the track in seconds
    #[serde(default)]
    pub track_length: u32,
    /// Musixmatch ID that is the same for equivalent tracks (e.g. same track on different albums)
    pub commontrack_id: u64,

    /// True if the track is instrumental
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub instrumental: bool,
    /// True if the lyrics contain explicit language
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub explicit: bool,
    /// True if lyrics are available
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub has_lyrics: bool,
    /// True if subtitles (synchronized lyrics) are available
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub has_subtitles: bool,
    /// True if richsync lyrics (synchronized by word) are available
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub has_richsync: bool,
    /// True if the track structure is available
    #[serde(default, deserialize_with = "crate::api_model::bool_from_int")]
    pub has_track_structure: bool,

    /// Amount of users that favorited the track on Musixmatch
    #[serde(default)]
    pub num_favourite: u32,

    /// Musixmatch lyrics ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_zero")]
    pub lyrics_id: Option<u64>,
    /// Musixmatch subtitle ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_zero")]
    pub subtitle_id: Option<u64>,

    /// Musixmatch album ID
    pub album_id: u64,
    /// Album name
    pub album_name: String,

    /// Musixmatch artist ID
    pub artist_id: u64,
    /// Artist name
    pub artist_name: String,
    /// Musicbrainz artist ID
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub artist_mbid: Option<String>,

    /// Album cover URL (100x100px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_100x100: Option<String>,
    /// Album cover URL (350x350px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_350x350: Option<String>,
    /// Album cover URL (500x500px)
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_500x500: Option<String>,
    /// Album cover URL (800x800px)
    ///
    /// **Note:** not present on a lot of albums
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub album_coverart_800x800: Option<String>,

    /// Human-readable Musixmatch ID
    ///
    /// Used in the URLs on the Musixmatch website.
    ///
    /// Example: `aespa/Black-Mamba`
    pub commontrack_vanity_id: String,

    /// Track release date
    #[serde(default, with = "crate::api_model::optional_datetime")]
    pub first_release_date: Option<OffsetDateTime>,
    /// Date and time when the track was last updated
    #[serde(with = "time::serde::rfc3339")]
    pub updated_time: OffsetDateTime,

    /// List of primary genres
    #[serde(default)]
    pub primary_genres: Genres,
    /// Secondary genres / Subgenres
    ///
    /// Example: primary_genres: Pop, secondary_genres: K-Pop / Mandopop
    ///
    /// Note that this schema is not applied to all tracks on Musixmatch.
    /// There are for example many K-Pop tracks with both Pop and K-Pop
    /// tagged as primary genre and this field empty.
    #[serde(default)]
    pub secondary_genres: Genres,

    /// Status of lyrics translation
    #[serde(default)]
    pub track_lyrics_translation_status: Vec<TrackLyricsTranslationStatus>,
}

/// Status of lyrics translation (language + progress)
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct TrackLyricsTranslationStatus {
    /// Source language code (e.g. "ko")
    pub from: Option<String>,
    /// Target language code (e.g. "en")
    pub to: String,
    /// Translation ratio from 0 (untranslated) - 1 (fully translated)
    ///
    /// **NOT** the percentage
    pub perc: f32,
}

/// Available track charts
#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum ChartName {
    /// Editorial chart
    Top,
    /// Most viewed lyrics in the last 2 hours
    Hot,
}

impl ChartName {
    pub(crate) fn as_str(&self) -> &str {
        match self {
            ChartName::Top => "top",
            ChartName::Hot => "hot",
        }
    }
}
