use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use time::OffsetDateTime;

use super::{subtitle::SubtitleLines, SubtitleLine};

#[derive(Debug, Deserialize)]
pub(crate) struct TranslationListBody {
    pub translations_list: Vec<TranslationBody>,
}

#[derive(Debug, Deserialize)]
pub(crate) struct TranslationBody {
    pub translation: Translation,
}

/// Translation: a translated line of lyrics
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Translation {
    /// source language code (e.g. "en")
    #[serde(default, deserialize_with = "crate::api_model::null_if_empty")]
    pub language_from: Option<String>,
    /// Translated line
    #[serde(default)]
    pub description: String,
    /// Original line
    #[serde(default)]
    pub matched_line: String,
    /// Date when the translation was created
    #[serde(with = "time::serde::rfc3339")]
    pub created_date: OffsetDateTime,
    /// Date when the translation was last updated
    #[serde(with = "time::serde::rfc3339")]
    pub last_updated: OffsetDateTime,
}

/// List of translated lines
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct TranslationList {
    /// List of translated lines
    pub lines: Vec<Translation>,
    /// 2-character language code (e.g. "en")
    pub lang: String,
}

/// [`HashMap`] of translated lines
///
/// TranslationMaps are used to translate the original lyrics or subtitles.
#[derive(Clone, Debug)]
pub struct TranslationMap {
    /// Translation map
    t_map: HashMap<String, String>,
    /// 2-character language code (e.g. "en")
    lang: String,
}

impl From<TranslationList> for TranslationMap {
    fn from(list: TranslationList) -> Self {
        TranslationMap {
            t_map: list
                .lines
                .into_iter()
                .map(|t| (t.matched_line, t.description))
                .collect(),
            lang: list.lang,
        }
    }
}

impl TranslationList {
    /// Get the number of translated lines
    pub fn len(&self) -> usize {
        self.lines.len()
    }

    /// Returns `true` if there are no translated lines.
    pub fn is_empty(&self) -> bool {
        self.lines.is_empty()
    }
}

impl TranslationMap {
    /// Get the number of translated lines
    pub fn len(&self) -> usize {
        self.t_map.len()
    }

    /// Returns `true` if there are no translated lines.
    pub fn is_empty(&self) -> bool {
        self.t_map.is_empty()
    }

    /// Translate the original lyrics fetched with [Musixmatch::track_lyrics](crate::Musixmatch::track_lyrics)
    pub fn translate_lyrics(&self, lyrics_body: &str) -> String {
        let mut t = lyrics_body
            .split('\n')
            .map(|line| {
                self.t_map
                    .get(line)
                    .map_or_else(|| line.to_owned(), |trans| trans.to_owned())
                    + "\n"
            })
            .collect::<String>();

        // Remove trailing newline
        t.pop();
        t
    }

    /// Translate the original subtitles fetched with [Musixmatch::track_subtitle](crate::Musixmatch::track_subtitle)
    pub fn translate_subtitles(&self, subtitles: &SubtitleLines) -> SubtitleLines {
        SubtitleLines {
            lines: subtitles
                .lines
                .iter()
                .map(|line| SubtitleLine {
                    text: self
                        .t_map
                        .get(&line.text)
                        .map_or_else(|| line.text.to_owned(), |trans| trans.to_owned()),
                    time: line.time,
                })
                .collect::<Vec<_>>(),
            lang: Some(self.lang.to_owned()),
            length: subtitles.length,
        }
    }
}
