//! Persistent session storage

use std::{
    fs,
    path::{Path, PathBuf},
};

use log::error;

pub(crate) const DEFAULT_PATH: &str = "musixmatch_session.json";

/// Storage backend for the Musixmatch client
///
/// Used to persist sessions.
pub trait SessionStorage: Sync + Send {
    /// Write a string to the storage
    fn write(&self, data: &str);
    /// Read a string from the storage.
    ///
    /// Return [`None`] if the storage is empty.
    fn read(&self) -> Option<String>;
}

/// Storage implementation that stores the session as a JSON file
pub struct FileStorage {
    path: PathBuf,
}

impl FileStorage {
    /// Create a new `FileStorage` that stores the session at the given location.
    pub fn new<P: AsRef<Path>>(path: P) -> Self {
        Self {
            path: path.as_ref().to_path_buf(),
        }
    }
}

impl Default for FileStorage {
    fn default() -> Self {
        Self {
            path: Path::new(DEFAULT_PATH).into(),
        }
    }
}

impl SessionStorage for FileStorage {
    fn write(&self, data: &str) {
        fs::write(&self.path, data).unwrap_or_else(|e| {
            error!(
                "Could not write session to file `{}`. Error: {}",
                self.path.to_string_lossy(),
                e
            );
        });
    }

    fn read(&self) -> Option<String> {
        if !self.path.exists() {
            return None;
        }

        match fs::read_to_string(&self.path) {
            Ok(data) => Some(data),
            Err(e) => {
                error!(
                    "Could not load session from file `{}`. Error: {}",
                    self.path.to_string_lossy(),
                    e
                );
                None
            }
        }
    }
}
